#!/usr/bin/env python

from __future__ import annotations

from pathlib import Path
from typing import Any, Collection

import pandas as pd
from datasets import Dataset, load_metric
from setfit import SetFitModel, SetFitTrainer

from .estimator_base import EstimatorBase

###############################################################################


class TransformerEstimator(EstimatorBase):

    # Properties
    trainer: SetFitTrainer | None = None

    def __init__(
        self: "TransformerEstimator",
        base_model: str = "distilbert-base-uncased",
        training_args: dict[str, Any] | None = None,
        eval_size: float = 0.2,
        output_dir: Path | str | None = None,
        verbose: bool = False,
        **kwargs: dict[str, Any],
    ) -> None:
        self.base_model = base_model
        self.eval_size = eval_size
        self.verbose = verbose

        # Handle output dir
        if output_dir:
            self.model_dir = Path(output_dir).resolve()
        else:
            self.model_dir = Path("lazy-text-setfit-transformer/").resolve()

        # Handle training arguments
        if training_args:
            self.training_args = training_args
        else:
            # Determine logging steps
            self.training_args = {
                "batch_size": 4,
                "num_epochs": 1,
            }

    def fit(
        self: "TransformerEstimator",
        x: Collection[str],
        y: Collection[str],
    ) -> "TransformerEstimator":
        """
        Fit the estimator.

        Parameters
        ----------
        x: Collection[str]
            The training data.
        y: Collection[str]
            The testing data.

        Returns
        -------
        "TransformerEstimator"
            The estimator.
        """
        # Remove printing
        if not self.verbose:
            from ..logging_utils import set_global_logging_level

            set_global_logging_level()

        # Create dataframes and split to eval
        df_all = pd.DataFrame(
            {
                "text": x,
                "label": y,
            }
        )

        # Make the label luts
        label_names = df_all.label.unique()
        label2id, id2label = {}, {}
        for i, label in enumerate(label_names):
            label2id[label] = str(i)
            id2label[str(i)] = label

        self.label2id = label2id
        self.id2label = id2label

        # Make the model
        model = SetFitModel.from_pretrained(self.base_model)

        # Create datasets
        train_dataset = Dataset.from_pandas(df_all)
        train_dataset = train_dataset.class_encode_column("label")

        # Load metrics and create metric compute func
        f1_metric = load_metric("f1")

        def compute_metrics(
            y_pred: Collection[str], y_test: Collection[str]
        ) -> dict | None:
            f1_score = f1_metric.compute(
                predictions=y_pred,
                references=y_test,
                average="weighted",
            )
            return f1_score

        # Determine number of training iterations
        if "num_iterations" not in self.training_args:
            # This is sort of a hack, but as you increase in dataset size,
            # you don't need as many training iterations
            self.training_args["num_iterations"] = max(200 // len(df_all), 4)

        # Create trainer
        self.trainer = SetFitTrainer(
            model=model,
            train_dataset=train_dataset,
            metric=compute_metrics,
            **self.training_args,
        )

        # Train
        self.trainer.train()
        self.trainer.model.save_pretrained(str(self.model_dir))

        # TODO: save label2id and id2label

        return self

    def predict(
        self: "TransformerEstimator",
        x: Collection[str],
    ) -> Collection[str]:
        """
        Predict the values using the fitted estimator.

        Parameters
        ----------
        x: Collection[str]
            The data to predict.

        Returns
        -------
        Collection[str]
            The predictions.
        """
        if self.trainer is not None:
            preds = self.trainer.model.predict(x)
            return [self.id2label[str(pred)] for pred in preds]

        raise ValueError("SetFit model has not been trained yet.")


def _make_pipeline(
    **kwargs: Any,
) -> TransformerEstimator:
    return TransformerEstimator(**kwargs)



#!/usr/bin/env python

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection

###############################################################################


class EstimatorBase(ABC):
    """Base class for estimators."""

    @abstractmethod
    def fit(
        self: "EstimatorBase",
        x: Collection[str],
        y: Collection[str],
    ) -> "EstimatorBase":
        """
        Fit the estimator.

        Parameters
        ----------
        x: Collection[str]
            The training data.
        y: Collection[str]
            The testing data.

        Returns
        -------
        "EstimatorBase"
            The estimator.
        """
        pass

    @abstractmethod
    def predict(
        self: "EstimatorBase",
        x: Collection[str],
    ) -> Collection[str]:
        """
        Predict the values using the fitted estimator.

        Parameters
        ----------
        x: Collection[str]
            The data to predict.

        Returns
        -------
        Collection[str]
            The predictions.
        """
        pass

    # TODO: save function



#!/usr/bin/env python

"""
Configuration for tests! There are a whole list of hooks you can define in this file to
run before, after, or to mutate how tests run. Commonly for most of our work, we use
this file to define top level fixtures that may be needed for tests throughout multiple
test files.

In this case, while we aren't using this fixture in our tests, the prime use case for
something like this would be when we want to preload a file to be used in multiple
tests. File reading can take time, so instead of re-reading the file for each test,
read the file once then use the loaded content.

Docs: https://docs.pytest.org/en/latest/example/simple.html
      https://docs.pytest.org/en/latest/plugins.html#requiring-loading-plugins-in-a-test-module-or-conftest-file
"""

from pathlib import Path

import pytest


@pytest.fixture
def data_dir() -> Path:
    return Path(__file__).parent / "data"



#!/usr/bin/env python

import logging
import re


def set_global_logging_level(
    level: int = logging.ERROR,
    prefices: list[str] | None = None,
) -> None:
    """
    Override logging levels of different modules based on their name as a prefix.
    It needs to be invoked after the modules have been loaded so that their
    loggers have been initialized.

    Parameters
    ----------
    level: int
        Desired level. e.g. logging.INFO.
        Default is logging.ERROR
    prefices: list[str] | None
        One or more string prefices to match (e.g. ["transformers", "torch"])
        Default of None will match all loggers.
        The match is a case-sensitive `module_name.startswith(prefix)`

    Notes
    -----
    Credit:
    https://github.com/huggingface/transformers/issues/3050#issuecomment-682167272
    """
    if prefices is None:
        prefices = [""]

    prefix_re = re.compile(rf"^(?:{ '|'.join(prefices) })")
    for name in logging.root.manager.loggerDict:
        if re.match(prefix_re, name):
            logging.getLogger(name).setLevel(level)



"""Top-level package for lazy_text_classifiers."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("lazy-text-classifiers")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Eva Maxfield Brown"
__email__ = "evamaxfieldbrown@gmail.com"

from .model_selection import LazyTextClassifiers  # noqa: F401



#!/usr/bin/env python

from __future__ import annotations

from typing import Any

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegressionCV
from sklearn.pipeline import Pipeline, make_pipeline

###############################################################################


def _make_pipeline(
    vectorizer_kwargs: dict[str, Any] | None = None,
    logit_regression_cv_kwargs: dict[str, Any] | None = None,
    verbose: bool = False,
    **kwargs: Any,
) -> Pipeline:
    # Make vectorizer
    if vectorizer_kwargs:
        vect = TfidfVectorizer(
            **vectorizer_kwargs,
        )
    else:
        vect = TfidfVectorizer(
            strip_accents="unicode",
            lowercase=True,
            stop_words="english",
        )

    if logit_regression_cv_kwargs:
        logit = LogisticRegressionCV(
            **logit_regression_cv_kwargs,
        )
    else:
        logit = LogisticRegressionCV()

    return make_pipeline(
        vect,
        logit,
        verbose=verbose,
    )



#!/usr/bin/env python

from __future__ import annotations

import random
import time
from typing import TYPE_CHECKING, Any, Collection

import numpy as np
import pandas as pd
import torch
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    precision_recall_fscore_support,
)

from .constants import MODEL_NAME_WRAPPER_LUT, ModelNames

if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline

    from .model_wrappers.estimator_base import EstimatorBase

###############################################################################


class LazyTextClassifiers:

    # TODO: optuna??

    def __init__(
        self: "LazyTextClassifiers",
        verbose: int = 1,
        ignore_warnings: bool = True,
        random_state: int | None = None,
    ) -> None:
        """
        Class for managing fitting all classifiers.

        Parameters
        ----------
        verbose: int
            Logging verbosity level.
            Levels:
                0 - no logging;
                1 - log basic progress;
                2 - log everything;
            Default: 1 (log basic progress)
        ignore_warnings: bool
            Should all warnings be ignores.
        random_state: int | None
            A seed to initialize random state.
            Default: None (no pre-set random seed)
        """
        self.verbose = verbose

        # Set up warnings
        if ignore_warnings:
            import warnings

            warnings.filterwarnings("ignore")

        # Handle seed
        if random_state:
            torch.manual_seed(random_state)
            random.seed(random_state)
            np.random.seed(random_state)

    def fit(
        self: "LazyTextClassifiers",
        x_train: Collection[str],
        x_test: Collection[str],
        y_train: Collection[str],
        y_test: Collection[str],
        model_kwargs: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """
        Fit all models with the provided data.

        Parameters
        ----------
        x_train: Collection[str]
            The training data, an iterable object where each item is a string.
        x_test: Collection[str]
            The testing data, an iterable object where each item is a string.
        y_train: Collection[str]
            The training labels, an iterable object where each item is a class.
        y_test: Collection[str]
            The testing labels, an iterable object where each item is a class.
        model_kwargs: dict[str, Any] | None
            Any specific model kwargs to pass through.
            Default None (use default parameters and settings for all models)

        Returns
        -------
        pd.DataFrame
            The results and metrics returned from fitting all models.
        """
        # Handle kwargs
        if model_kwargs is None:
            model_kwargs = {}

        # Remove setfit model if data is larger than 200
        if len(x_train) > 200:
            MODEL_NAME_WRAPPER_LUT.pop(ModelNames.setfit_transformer)

        # Iterate model fitting and preds
        self.fit_models: dict[str, "EstimatorBase" | "Pipeline"] = {}
        results_rows: list[dict[str, str | float]] = []
        for model_name, model_wrapper in MODEL_NAME_WRAPPER_LUT.items():
            # Start perf time
            start_time = time.perf_counter()

            # Instantiate estimate with any extra kwargs
            if self.verbose > 0:
                print(f"Initializing model: '{model_name}'")
            estimator = model_wrapper(
                **model_kwargs.get(model_name, {}),
                verbose=self.verbose > 1,
            )

            # Run the fit method passing in x_train and y_train
            if self.verbose > 0:
                print(f"Fitting model: '{model_name}'")
            estimator = estimator.fit(x_train, y_train)
            self.fit_models[model_name] = estimator

            # Record training time
            duration = time.perf_counter() - start_time

            # Run the eval
            if self.verbose > 0:
                print(f"Testing model: '{model_name}'")

            # Handle pd.Series
            if isinstance(x_test, pd.Series):
                x_test = x_test.tolist()
            preds = estimator.predict(x_test)

            # Calc metrics
            acc = accuracy_score(
                y_test,
                preds,
            )
            bal_acc = balanced_accuracy_score(
                y_test,
                preds,
            )
            pre, rec, f1, _ = precision_recall_fscore_support(
                y_test,
                preds,
                average="weighted",
            )

            # Combine and log if desired
            this_model_result = {
                "model": model_name,
                "accuracy": acc,
                "balanced_accuracy": bal_acc,
                "precision": pre,
                "recall": rec,
                "f1": f1,
                "time": duration,
            }
            if self.verbose > 0:
                print(f"'{model_name}' eval results: {this_model_result}")

            results_rows.append(this_model_result)

        # Create dataframe of results
        self.results_df = (
            pd.DataFrame(
                results_rows,
            )
            .sort_values(by="f1", ascending=False)
            .reset_index(drop=True)
        )

        return self.results_df



#!/usr/bin/env python

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Callable

from .model_wrappers import (
    fine_tuned_transformer,
    semantic_logit,
    setfit_transformer,
    tfidf_logit,
)

if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline

    from .model_wrappers.estimator_base import EstimatorBase


class ModelNames:
    tfidf_logit = "tfidf-logit"
    semantic_logit = "semantic-logit"
    setfit_transformer = "setfit-transformer"
    fine_tuned_transformer = "fine-tuned-transformer"


# Use a varienty of base models for the semantic models
SEMANTIC_BASE_MODELS = {
    "distilbert-sst2": "distilbert-base-uncased-finetuned-sst-2-english",
    "distilbert-emotions": "joeddav/distilbert-base-uncased-go-emotions-student",
}


SEMANTIC_MODEL_VARIANTS = {}
for short_base_model_name, full_base_model_name in SEMANTIC_BASE_MODELS.items():
    SEMANTIC_MODEL_VARIANTS[f"semantic-logit-{short_base_model_name}"] = partial(
        semantic_logit._make_pipeline,
        sentence_encoder_kwargs={
            "name": full_base_model_name,
        },
    )
    SEMANTIC_MODEL_VARIANTS[f"fine-tuned-{short_base_model_name}"] = partial(
        fine_tuned_transformer._make_pipeline,
        base_model=full_base_model_name,
    )


MODEL_NAME_WRAPPER_LUT: dict[str, Callable[..., "Pipeline" | "EstimatorBase"]] = {
    ModelNames.tfidf_logit: tfidf_logit._make_pipeline,
    **SEMANTIC_MODEL_VARIANTS,
    ModelNames.setfit_transformer: setfit_transformer._make_pipeline,
}



#!/usr/bin/env python

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Collection

import numpy as np
import pandas as pd
from datasets import Dataset, load_metric
from sklearn.model_selection import train_test_split
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    DataCollatorWithPadding,
    EvalPrediction,
    Trainer,
    TrainingArguments,
    pipeline,
)
from transformers.trainer_callback import PrinterCallback

from .estimator_base import EstimatorBase

if TYPE_CHECKING:
    from datasets.arrow_dataset import Batch
    from transformers.tokenization_utils_base import BatchEncoding

###############################################################################


class TransformerEstimator(EstimatorBase):

    # Properties
    trainer: Trainer | None = None

    def __init__(
        self: "TransformerEstimator",
        base_model: str = "distilbert-base-uncased",
        training_args: TrainingArguments | None = None,
        eval_size: float = 0.2,
        output_dir: Path | str | None = None,
        verbose: bool = False,
        **kwargs: dict[str, Any],
    ) -> None:
        self.base_model = base_model
        self.eval_size = eval_size
        self.verbose = verbose

        # Handle output dir
        if output_dir:
            self.model_dir = Path(output_dir).resolve()
        else:
            self.model_dir = Path(f"fine-tuned-{base_model}").resolve()

        # Handle training arguments
        if training_args:
            self.training_args = training_args
        else:
            # Determine logging steps
            logging_args = {}
            if verbose:
                logging_args["logging_steps"] = 10

            self.training_args = TrainingArguments(
                output_dir=self.model_dir,
                evaluation_strategy="epoch",
                save_strategy="epoch",
                learning_rate=3e-5,
                load_best_model_at_end=True,
                metric_for_best_model="f1",
                per_device_eval_batch_size=4,
                per_device_train_batch_size=4,
                num_train_epochs=5,
                weight_decay=0.01,
                **logging_args,
            )

    def fit(
        self: "TransformerEstimator",
        x: Collection[str],
        y: Collection[str],
    ) -> "TransformerEstimator":
        """
        Fit the estimator.

        Parameters
        ----------
        x: Collection[str]
            The training data.
        y: Collection[str]
            The testing data.

        Returns
        -------
        "TransformerEstimator"
            The estimator.
        """
        # Remove printing
        if not self.verbose:
            from ..logging_utils import set_global_logging_level

            set_global_logging_level()

        # Create dataframes and split to eval
        df_all = pd.DataFrame(
            {
                "text": x,
                "label": y,
            }
        )
        train_df, eval_df = train_test_split(
            df_all,
            test_size=self.eval_size,
            stratify=df_all["label"],
        )

        # Make the label luts
        label_names = df_all.label.unique()
        label2id, id2label = {}, {}
        for i, label in enumerate(label_names):
            label2id[label] = str(i)
            id2label[str(i)] = label

        # Make the model
        model = AutoModelForSequenceClassification.from_pretrained(
            pretrained_model_name_or_path=self.base_model,
            num_labels=len(id2label),
            label2id=label2id,
            id2label=id2label,
            ignore_mismatched_sizes=True,
        )

        # Create datasets
        train_dataset = Dataset.from_pandas(train_df)
        eval_dataset = Dataset.from_pandas(eval_df)
        train_dataset = train_dataset.class_encode_column("label")
        eval_dataset = eval_dataset.class_encode_column("label")

        # Create tokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.base_model)

        def preprocess_function(examples: "BatchEncoding") -> "Batch":
            return tokenizer(examples["text"], truncation=True)

        # Preprocess data
        tokenized_train_dataset = train_dataset.map(preprocess_function, batched=True)
        tokenized_eval_dataset = eval_dataset.map(preprocess_function, batched=True)

        # Load metrics and create metric compute func
        f1_metric = load_metric("f1")

        def compute_metrics(eval_pred: EvalPrediction) -> dict | None:
            predictions = np.argmax(eval_pred.predictions, axis=-1)
            f1_score = f1_metric.compute(
                predictions=predictions,
                references=eval_pred.label_ids,
                average="weighted",
            )
            return f1_score

        # Create data collator
        data_collator = DataCollatorWithPadding(tokenizer=tokenizer)

        # Create trainer
        self.trainer = Trainer(
            model=model,
            args=self.training_args,
            train_dataset=tokenized_train_dataset,
            eval_dataset=tokenized_eval_dataset,
            tokenizer=tokenizer,
            data_collator=data_collator,
            compute_metrics=compute_metrics,
        )

        # Remove printer callback if it exists
        if not self.verbose:
            self.trainer.remove_callback(PrinterCallback)

        # Train
        self.trainer.train()
        self.trainer.save_model()

        return self

    def predict(
        self: "TransformerEstimator",
        x: Collection[str],
    ) -> Collection[str]:
        """
        Predict the values using the fitted estimator.

        Parameters
        ----------
        x: Collection[str]
            The data to predict.

        Returns
        -------
        Collection[str]
            The predictions.
        """
        pipe = pipeline(
            "text-classification",
            model=str(self.model_dir),
            tokenizer=str(self.model_dir),
        )
        return [pred[0]["label"] for pred in pipe(x, truncation=True, top_k=1)]


def _make_pipeline(
    **kwargs: Any,
) -> TransformerEstimator:
    return TransformerEstimator(**kwargs)
